"""Python client for LGHorizon."""
# flake8: noqa
# Box states
ONLINE_RUNNING = "ONLINE_RUNNING"
ONLINE_STANDBY = "ONLINE_STANDBY"
UNKNOWN = "UNKNOWN"

BOX_PLAY_STATE_CHANNEL = "linear"
BOX_PLAY_STATE_REPLAY = "replay"
BOX_PLAY_STATE_DVR = "nDVR"
BOX_PLAY_STATE_BUFFER = "reviewbuffer"
BOX_PLAY_STATE_APP = "app"
BOX_PLAY_STATE_VOD = "VOD"

# List with available media keys.
MEDIA_KEY_POWER = "Power"
MEDIA_KEY_ENTER = "Enter"
MEDIA_KEY_ESCAPE = "Escape"  # Not yet implemented

MEDIA_KEY_HELP = "Help"  # Not yet implemented
MEDIA_KEY_INFO = "Info"  # Not yet implemented
MEDIA_KEY_GUIDE = "Guide"  # Not yet implemented

MEDIA_KEY_CONTEXT_MENU = "ContextMenu"  # Not yet implemented
MEDIA_KEY_CHANNEL_UP = "ChannelUp"
MEDIA_KEY_CHANNEL_DOWN = "ChannelDown"

MEDIA_KEY_RECORD = "MediaRecord"
MEDIA_KEY_PLAY_PAUSE = "MediaPlayPause"
MEDIA_KEY_STOP = "MediaStop"
MEDIA_KEY_REWIND = "MediaRewind"
MEDIA_KEY_FAST_FORWARD = "MediaFastForward"

RECORDING_TYPE_SINGLE = "single"
RECORDING_TYPE_SHOW = "show"
RECORDING_TYPE_SEASON = "season"

BE_AUTH_URL = "https://login.prd.telenet.be/openid/login.do"

COUNTRY_SETTINGS = {
    "nl": {
        "api_url": "https://prod.spark.ziggogo.tv",
        "personalization_url_format": "https://prod.spark.ziggogo.tv/nld/web/personalization-service/v1/customer/{household_id}/devices",
        "mqtt_url": "obomsg.prod.nl.horizon.tv",
        "use_oauth": False,
        "channels": [
            {
                "channelId": "NL_000073_019506",
                "channelName": "Netflix",
                "channelNumber": "150",
            },
            {
                "channelId": "NL_000074_019507",
                "channelName": "Videoland",
                "channelNumber": "151",
            },
            {
                "channelId": "NL_000194_019352",
                "channelName": "NPO",
                "channelNumber": "152"
            },
            {
                "channelId": "NL_000199_019356",
                "channelName": "Prime Video",
                "channelNumber": "153"
            }
        ],
        "platform_types":{
            "EOS":{
                "manufacturer": "Arris",
                "model": "DCX960"
            },
            "APOLLO":{
                "manufacturer": "Arris",
                "model": "VIP5002W"
            }
        },
        "language":"nl"
    },
    "ch": {
        "api_url": "https://prod.spark.sunrisetv.ch",
        "personalization_url_format": "https://prod.spark.sunrisetv.ch/eng/web/personalization-service/v1/customer/{householdId}/devices",
        "mqtt_url": "obomsg.prod.ch.horizon.tv",
        "use_oauth": False,
        "channels": [],
        "language":"de"
    },
    "be-nl": {
        "api_url": "https://prod.spark.telenet.tv",
        "personalization_url_format": "https://prod.spark.telenettv.be/nld/web/personalization-service/v1/customer/{household_id}/devices",
        "mqtt_url": "obomsg.prod.be.horizon.tv",
        "use_oauth": True,
        "oauth_username_fieldname": "j_username",
        "oauth_password_fieldname": "j_password",
        "oauth_add_accept_header": False,
        "oauth_url": "https://login.prd.telenet.be/openid/login.do",
        "oauth_quote_login": False,
        "oauth_redirect_header": "Location",
        "channels": [
            {"channelId": "netflix", "channelName": "Netflix", "channelNumber": "600"},
            {"channelId": "youtube", "channelName": "Youtube", "channelNumber": "-1"},
        ],
        "platform_types":{
            "EOS":{
                "manufacturer": "Arris",
                "model": "DCX960"
            },
            "EOS2":{
                "manufacturer": "HUMAX",
                "model": "2008C-STB-TN"
            }
        },
        "language":"nl"
    },
    # "be-nl-preprod": {
    #     "api_url": "https://web-api-preprod-obo.horizon.tv/oesp/v4/BE/nld/web",
    #     "personalization_url_format": "https://preprod.spark.telenettv.be/nld/web/personalization-service/v1/customer/{household_id}/devices",
    #     "mqtt_url": "obomsg.preprod.be.horizon.tv",
    #     "use_oauth": True,
    #     "oauth_username_fieldname": "j_username",
    #     "oauth_password_fieldname": "j_password",
    #     "oauth_add_accept_header": False,
    #     "oauth_url": "https://login.prd.telenet.be/openid/login.do",
    #     "oauth_quote_login": False,
    #     "oauth_redirect_header": "Location",
    #     "channels": [
    #         {"channelId": "netflix", "channelName": "Netflix", "channelNumber": "600"},
    #         {"channelId": "youtube", "channelName": "Youtube", "channelNumber": "-1"},
    #     ],
    # },
    # "be-fr": {
    #     "api_url": "https://web-api-prod-obo.horizon.tv/oesp/v4/BE/fr/web",
    #     "personalization_url_format": "https://prod.spark.telenettv.be/fr/web/personalization-service/v1/customer/{household_id}/devices",
    #     "mqtt_url": "obomsg.prod.be.horizon.tv",
    #     "use_oauth": True,
    #     "oauth_username_fieldname": "j_username",
    #     "oauth_password_fieldname": "j_password",
    #     "oauth_add_accept_header": False,
    #     "oauth_url": "https://login.prd.telenet.be/openid/login.do",
    #     "oauth_quote_login": False,
    #     "oauth_redirect_header": "Location",
    #     "channels": [],
    # },
    "at": {
        "api_url": "https://prod.spark.magentatv.at",
        "personalization_url_format": "https://prod.spark.magentatv.at/deu/web/personalization-service/v1/customer/{householdId}/devices",
        "mqtt_url": "obomsg.prod.at.horizon.tv",
        "use_oauth": False,
        "channels": [],
        "language":"de"
    },
    # "gb": {
    #     "api_url": "https://web-api-prod-obo.horizon.tv/oesp/v4/GB/eng/web/",
    #     "personalization_url_format": "https://prod.spark.virginmedia.com/eng/web/personalization-service/v1/customer/{household_id}/devices",
    #     "mqtt_url": "obomsg.prod.gb.horizon.tv",
    #     "use_oauth": True,
    #     "oauth_username_fieldname": "username",
    #     "oauth_password_fieldname": "credential",
    #     "oauth_add_accept_header": True,
    #     "oauth_url": "https://id.virginmedia.com/rest/v40/session/start?protocol=oidc&rememberMe=true",
    #     "oauth_quote_login": True,
    #     "oauth_redirect_header": "x-redirect-location",
    #     "channels": [],
    # },
    # "gb-preprod": {
    #     "api_url": "https://web-api-preprod-obo.horizon.tv/oesp/v4/GB/eng/web/",
    #     "personalization_url_format": "https://preprod.spark.virginmedia.com/eng/web/personalization-service/v1/customer/{household_id}/devices",
    #     "mqtt_url": "obomsg.preprod.gb.horizon.tv",
    #     "use_oauth": False,
    #     "oauth_username_fieldname": "username",
    #     "oauth_password_fieldname": "credential",
    #     "oauth_add_accept_header": True,
    #     "oauth_url": "https://id.virginmedia.com/rest/v40/session/start?protocol=oidc&rememberMe=true",
    #     "oauth_quote_login": True,
    #     "oauth_redirect_header": "x-redirect-location",
    #     "channels": [],
    # },
    "ie": {
        "api_url": "https://prod.spark.virginmediatv.ie",
        "personalization_url_format": "https://prod.spark.virginmediatv.ie/eng/web/personalization-service/v1/customer/{householdId}/devices",
        "mqtt_url": "obomsg.prod.ie.horizon.tv",
        "use_oauth": False,
        "channels": [],
        "language":"en"
    },
    # "pl": {
    #     "api_url": "https://web-api-pepper.horizon.tv/oesp/v4/PL/pol/web",
    #     "personalization_url_format": "https://prod.spark.unknown.pl/pol/web/personalization-service/v1/customer/{householdId}/devices",
    #     "mqtt_url": "obomsg.prod.pl.horizon.tv/mqtt",
    #     "use_oauth": False,
    #     "channels": [],
    # },
    # "hu": {
    #     "api_url": "https://web-api-pepper.horizon.tv/oesp/v4/HU/HUN/web",
    #     "personalization_url_format": "https://prod.spark.upctv.hu/HUN/web/personalization-service/v1/customer/{householdId}/devices",  # guessed
    #     "mqtt_url": "obomsg.prod.hu.horizon.tv/mqtt",
    #     "use_oauth": False,
    #     "channels": [],
    # },
    # "de": {
    #     "api_url": "https://web-api-pepper.horizon.tv/oesp/v4/DE/DEU/web",
    #     "personalization_url_format": "https://prod.spark.upctv.de/DEU/web/personalization-service/v1/customer/{householdId}/devices",  # guessed
    #     "mqtt_url": "obomsg.prod.de.horizon.tv/mqtt",
    #     "use_oauth": False,
    #     "channels": [],
    # },
    # "cz": {
    #     "api_url": "https://web-api-pepper.horizon.tv/oesp/v4/CZ/ces/web/",
    #     "personalization_url_format": "https://prod.spark.upctv.cz/ces/web/personalization-service/v1/customer/{householdId}/devices",  # guessed
    #     "mqtt_url": "obomsg.prod.cz.horizon.tv/mqtt",
    #     "use_oauth": False,
    #     "channels": [],
    # },
    # "sk": {
    #     "api_url": "https://web-api-pepper.horizon.tv/oesp/v4/sk/slk/web/",
    #     "personalization_url_format": "https://prod.spark.upctv.sk/SLK/web/personalization-service/v1/customer/{householdId}/devices",  # guessed
    #     "mqtt_url": "obomsg.prod.sk.horizon.tv/mqtt",
    #     "use_oauth": False,
    #     "channels": [],
    # },
    # "ro": {
    #     "api_url": "https://web-api-pepper.horizon.tv/oesp/v4/ro/ron/web/",
    #     "personalization_url_format": "https://prod.spark.upctv.ro/ron/web/personalization-service/v1/customer/{householdId}/devices",  # guessed
    #     "mqtt_url": "obomsg.prod.ro.horizon.tv/mqtt",
    #     "use_oauth": False,
    #     "channels": [],
    # },
}



"""Python client for LG Horizon."""
from .lghorizon_api import LGHorizonApi
from .models import LGHorizonBox, LGHorizonRecordingListSeasonShow, LGHorizonRecordingSingle, LGHorizonRecordingShow, LGHorizonRecordingEpisode
from .exceptions import LGHorizonApiUnauthorizedError, LGHorizonApiConnectionError
from .const import ONLINE_RUNNING, ONLINE_STANDBY, RECORDING_TYPE_SHOW, RECORDING_TYPE_SEASON, RECORDING_TYPE_SINGLE# noqa



"""Python client for LGHorizon."""
import logging
import json
import sys, traceback
from .exceptions import LGHorizonApiUnauthorizedError, LGHorizonApiConnectionError
import backoff
from requests import Session, exceptions as request_exceptions
import re
from .models import (
    LGHorizonAuth, 
    LGHorizonBox,
    LGHorizonMqttClient,
    LGHorizonCustomer,
    LGHorizonChannel,
    LGHorizonReplayEvent,
    LGHorizonRecordingSingle,
    LGHorizonVod,
    LGHorizonApp,
    LGHorizonBaseRecording,
    LGHorizonRecordingListSeasonShow,
    LGHorizonRecordingEpisode,
    LGHorizonRecordingShow
    )

from .const import (
    COUNTRY_SETTINGS,
    BOX_PLAY_STATE_BUFFER,
    BOX_PLAY_STATE_CHANNEL,
    BOX_PLAY_STATE_DVR,
    BOX_PLAY_STATE_REPLAY,
    BOX_PLAY_STATE_VOD,
    RECORDING_TYPE_SINGLE,
    RECORDING_TYPE_SEASON,
    RECORDING_TYPE_SHOW)
from typing import Any, Dict, List

_logger = logging.getLogger(__name__)
_supported_platforms = ["EOS", "EOS2", "HORIZON", "APOLLO"]

class LGHorizonApi:
    """Main class for handling connections with LGHorizon Settop boxes."""

    _auth: LGHorizonAuth = LGHorizonAuth()
    _session: Session = None
    settop_boxes: Dict[str, LGHorizonBox] = {}
    _customer: LGHorizonCustomer = None
    _mqttClient: LGHorizonMqttClient = None
    _channels: Dict[str, LGHorizonChannel] = {}
    _country_settings = None
    recording_capacity:int = None

    def __init__(self, username: str, password: str, country_code: str = "nl") -> None:
        """Create LGHorizon API."""
        self.username = username
        self.password = password
        self._session = Session()
        self._country_settings = COUNTRY_SETTINGS[country_code]

    @backoff.on_exception(backoff.expo, LGHorizonApiConnectionError, max_tries=3, logger=_logger)
    def _authorize(self) -> None:
        if self._country_settings["use_oauth"]:
            self.authorize_sso()
        else:
            self._authorize_default()

    def _authorize_default(self) -> None:
        _logger.debug("Authorizing")
        auth_url = f"{self._country_settings['api_url']}/auth-service/v1/authorization"
        auth_headers = {
            "x-device-code": "web"
        }
        auth_payload = {
            "password": self.password,
            "username": self.username
        }
        try:
            auth_response = self._session.post(auth_url, headers=auth_headers, json=auth_payload)
        except Exception as ex:
            raise LGHorizonApiConnectionError("Unknown connection failure") from ex
            
        if not auth_response.ok:
            error_json = auth_response.json()
            error = error_json["error"]
            if error and error["statusCode"] == 97401:
                raise LGHorizonApiUnauthorizedError("Invalid credentials")
            elif error:
                raise LGHorizonApiConnectionError(error["message"])
            else:
                raise LGHorizonApiConnectionError("Unknown connection error")

        self._auth.fill(auth_response.json())
        _logger.debug("Authorization succeeded")

    def authorize_sso(self):

        try:
            login_session = Session()
            #Step 1 - Get Authorization data
            _logger.debug("Step 1 - Get Authorization data")
            auth_url = f"{self._country_settings['api_url']}/auth-service/v1/sso/authorization"
            auth_response = login_session.get(auth_url)
            if not auth_response.ok:
                raise LGHorizonApiConnectionError("Can't connect to authorization URL")
            auth_response_json = auth_response.json()
            authorizationUri = auth_response_json["authorizationUri"]
            authValidtyToken = auth_response_json["validityToken"]

            #Step 2 - Get Authorization cookie
            _logger.debug("Step 2 - Get Authorization cookie")

            auth_cookie_response = login_session.get(authorizationUri)
            if not auth_cookie_response.ok:
                raise LGHorizonApiConnectionError("Can't connect to authorization URL")
            
            _logger.debug("Step 3 - Login")

            username_fieldname = self._country_settings["oauth_username_fieldname"]
            pasword_fieldname = self._country_settings["oauth_password_fieldname"]

            payload = {
                username_fieldname: self.username,
                pasword_fieldname: self.password,
                "rememberme": 'true'
            }

            
            login_response = login_session.post(
                self._country_settings["oauth_url"], payload, allow_redirects=False
            )
            if not login_response.ok:
                raise LGHorizonApiConnectionError("Can't connect to authorization URL")
            redirect_url = login_response.headers[self._country_settings["oauth_redirect_header"]]
            
            redirect_response = login_session.get(redirect_url, allow_redirects=False)
            success_url = redirect_response.headers[self._country_settings["oauth_redirect_header"]]
            codeMatches = re.findall(r"code=(.*)&", success_url)
            
            authorizationCode = codeMatches[0]

            new_payload = {
                "authorizationGrant":{
                    "authorizationCode":authorizationCode,
                    "validityToken":authValidtyToken
                }
            }
            headers = {
                "content-type":"application/json",
            }
            post_result = login_session.post(auth_url, json.dumps(new_payload), headers = headers)
            self._auth.fill(post_result.json())
            self._session.cookies["ACCESSTOKEN"] = self._auth.accessToken
        except Exception as ex:
            pass 

    def _obtain_mqtt_token(self):
        _logger.debug("Obtain mqtt token...")
        mqtt_response = self._do_api_call(f"{self._country_settings['api_url']}/auth-service/v1/mqtt/token")
        self._auth.mqttToken = mqtt_response["token"]
        _logger.debug(f"MQTT token: {self._auth.mqttToken}")

    def connect(self) -> None:
        _logger.debug("Connect to API")
        self._authorize()
        self._obtain_mqtt_token()
        self._mqttClient = LGHorizonMqttClient(self._auth, self._country_settings, self._on_mqtt_connected, self._on_mqtt_message)
        self._register_customer_and_boxes()
        self._mqttClient.connect()

    def disconnect(self):
        """Disconnect."""
        _logger.debug("Disconnect from API")
        if not self._mqttClient.is_connected:
            return
        self._mqttClient.disconnect()

    def _on_mqtt_connected(self) -> None:
        _logger.debug("Connected to MQTT server. Registering all boxes...")
        box:LGHorizonBox
        for box in self.settop_boxes.values():
            box.register_mqtt()

    def _on_mqtt_message(self, message:str)-> None:
        if "source" in message:
            deviceId = message["source"]
            if not deviceId in self.settop_boxes.keys():
                return
            try:
                if "deviceType" in message and message["deviceType"] == "STB":
                    self.settop_boxes[deviceId].update_state(message)
                if "status" in message:
                    self._handle_box_update(deviceId, message)
            except Exception as ex:
                _logger.error("Could not handle status message")
                exc_type, exc_value, exc_traceback = sys.exc_info()
                _logger.error(f"Full message: {str(message)}")
                _logger.error(repr(traceback.format_exception(exc_value)))
                self.settop_boxes[deviceId].playing_info.reset()
                self.settop_boxes[deviceId].playing_info.set_paused(False)

    def _handle_box_update(self, deviceId:str, raw_message:Any) -> None:
        statusPayload = raw_message["status"]
        if "uiStatus" not in statusPayload:
            return
        uiStatus = statusPayload["uiStatus"]
        if uiStatus == "mainUI":
            playerState = statusPayload["playerState"]
            if "sourceType" not in playerState or "source" not in playerState:
                return
            source_type = playerState["sourceType"]
            state_source = playerState["source"]
            self.settop_boxes[deviceId].playing_info.set_paused(playerState["speed"] == 0)
            if source_type in (
                BOX_PLAY_STATE_CHANNEL,
                BOX_PLAY_STATE_BUFFER,
                BOX_PLAY_STATE_REPLAY
                ):
                eventId = state_source["eventId"]
                raw_replay_event = self._do_api_call(f"{self._country_settings['api_url']}/eng/web/linear-service/v2/replayEvent/{eventId}?returnLinearContent=true&language={self._country_settings['language']}")
                replayEvent = LGHorizonReplayEvent(raw_replay_event)
                channel = self._channels[replayEvent.channelId]
                self.settop_boxes[deviceId].update_with_replay_event(source_type, replayEvent, channel)
            elif source_type == BOX_PLAY_STATE_DVR:
                recordingId = state_source["recordingId"]
                raw_recording = self._do_api_call(f"{self._country_settings['api_url']}/eng/web/recording-service/customers/{self._auth.householdId}/details/single/{recordingId}?profileId=4504e28d-c1cb-4284-810b-f5eaab06f034&language={self._country_settings['language']}")
                recording = LGHorizonRecordingSingle(raw_recording)
                channel = self._channels[recording.channelId]
                self.settop_boxes[deviceId].update_with_recording(source_type, recording, channel)
            elif source_type == BOX_PLAY_STATE_VOD:
                titleId = state_source["titleId"]
                raw_vod = self._do_api_call(f"{self._country_settings['api_url']}/eng/web/vod-service/v2/detailscreen/{titleId}?language={self._country_settings['language']}&profileId=4504e28d-c1cb-4284-810b-f5eaab06f034&cityId={self._customer.cityId}")
                vod = LGHorizonVod(raw_vod)
                self.settop_boxes[deviceId].update_with_vod(source_type, vod)
        elif uiStatus == "apps":
            app = LGHorizonApp(statusPayload["appsState"])
            self.settop_boxes[deviceId].update_with_app('app', app)     

    @backoff.on_exception(backoff.expo, LGHorizonApiConnectionError, max_tries=3, logger=_logger)
    def _do_api_call(self, url:str, tries:int = 0) -> str:
        _logger.info(f"Executing API call to {url}")
        try:
            api_response = self._session.get(url)
            api_response.raise_for_status()
        except request_exceptions.HTTPError as httpEx:
            self._authorize()
            raise LGHorizonApiConnectionError(f"Unable to call {url}. Error:{str(httpEx)}")
        except Exception as ex:
            self._authorize()
            raise LGHorizonApiConnectionError(f"Unable to call {url}. Error:{str(ex)}")
        json_response = api_response.json()
        _logger.debug(f"Result API call: {json_response}")
        return json_response
       

    def _register_customer_and_boxes(self):
        _logger.info("Get personalisation info...")
        personalisation_result = self._do_api_call(f"{self._country_settings['api_url']}/eng/web/personalization-service/v1/customer/{self._auth.householdId}?with=profiles%2Cdevices")
        _logger.debug(f"Personalisation result: {personalisation_result}")
        self._customer = LGHorizonCustomer(personalisation_result)
        self._get_channels()
        if not "assignedDevices" in personalisation_result:
            _logger.warning("No boxes found.")
            return
        _logger.info("Registering boxes")
        for device in personalisation_result["assignedDevices"]:
            platform_type = device["platformType"]
            if not platform_type in _supported_platforms:
                continue
            if "platform_types" in self._country_settings and platform_type in self._country_settings["platform_types"]:
                platformType = self._country_settings["platform_types"][platform_type]
            else:
                platformType = None
            box = LGHorizonBox(device, platformType, self._mqttClient,self._auth, self._channels)
            self.settop_boxes[box.deviceId] = box
            _logger.info(f"Box {box.deviceId} registered...")
            
    def _get_channels(self):
        _logger.info("Retrieving channels...")
        channels_result = self._do_api_call(f"{self._country_settings['api_url']}/eng/web/linear-service/v2/channels?cityId={self._customer.cityId}&language={self._country_settings['language']}&productClass=Orion-DASH")
        for channel in channels_result:
            channel_id = channel["id"]
            self._channels[channel_id] = LGHorizonChannel(channel)
        _logger.info(f"{len(self._channels)} retrieved.")

    def _get_replay_event(self, listingId) -> Any: 
        """Get listing."""
        _logger.info("Retrieving replay event details...")
        response = self._do_api_call(f"{self._country_settings['api_url']}/eng/web/linear-service/v2/replayEvent/{listingId}?returnLinearContent=true&language={self._country_settings['language']}")
        _logger.info("Replay event details retrieved")
        return response

    def  get_recording_capacity(self) -> int:
        """Returns remaining recording capacity"""
        try:
            _logger.info("Retrieving recordingcapacity...")
            quota_content = self._do_api_call(f"{self._country_settings['api_url']}/eng/web/recording-service/customers/{self._auth.householdId}/quota")
            if not "quota" in quota_content and not "occupied" in quota_content:
                _logger.error("Unable to fetch recording capacity...")
                return None
            capacity =  (quota_content["occupied"] / quota_content["quota"]) * 100
            self.recording_capacity = round(capacity)
            _logger.debug(f"Remaining recordingcapacity {self.recording_capacity}%")
            return self.recording_capacity
        except:
            _logger.error("Unable to fetch recording capacity...")
            return None
    
    def get_recordings(self) -> List[LGHorizonBaseRecording]:
        _logger.info("Retrieving recordings...")
        recording_content = self._do_api_call(f"{self._country_settings['api_url']}/eng/web/recording-service/customers/{self._auth.householdId}/recordings?sort=time&sortOrder=desc&language={self._country_settings['language']}")
        recordings= []
        for recording_data_item in recording_content['data']:
            type = recording_data_item['type']
            if type == RECORDING_TYPE_SINGLE:
                recordings.append(LGHorizonRecordingSingle(recording_data_item))
            elif type in (RECORDING_TYPE_SEASON, RECORDING_TYPE_SHOW):
                recordings.append(LGHorizonRecordingListSeasonShow(recording_data_item))
        _logger.info(F"{len(recordings)} recordings retrieved...")
        return recordings

    def get_recording_show(self, showId:str) -> list[LGHorizonRecordingSingle]:
        _logger.info("Retrieving show recordings...")
        show_recording_content = self._do_api_call(f"{self._country_settings['api_url']}/eng/web/recording-service/customers/{self._auth.householdId}/episodes/shows/{showId}?source=recording&language=nl&sort=time&sortOrder=asc")
        recordings = []
        for item in show_recording_content["data"]:
            if item['source'] == 'show':
                recordings.append(LGHorizonRecordingShow(item))
            else:
                recordings.append(LGHorizonRecordingEpisode(item))
        _logger.info(F"{len(recordings)} showrecordings retrieved...")
        return recordings


from datetime import datetime
from typing import Any, Callable, Dict, List
import paho.mqtt.client as mqtt
import requests
from .const import (
    BOX_PLAY_STATE_BUFFER,
    BOX_PLAY_STATE_CHANNEL,
    BOX_PLAY_STATE_DVR,
    BOX_PLAY_STATE_REPLAY,
    BOX_PLAY_STATE_APP,
    BOX_PLAY_STATE_VOD,
    ONLINE_STANDBY,
    ONLINE_RUNNING,
    MEDIA_KEY_POWER,
    MEDIA_KEY_PLAY_PAUSE,
    MEDIA_KEY_STOP,
    MEDIA_KEY_CHANNEL_UP,
    MEDIA_KEY_CHANNEL_DOWN,
    MEDIA_KEY_ENTER,
    MEDIA_KEY_REWIND,
    MEDIA_KEY_FAST_FORWARD,
    MEDIA_KEY_RECORD,
    RECORDING_TYPE_SEASON,
    RECORDING_TYPE_SHOW
)

import json
from .helpers import make_id
import logging

_logger = logging.getLogger(__name__)

class LGHorizonAuth:
    householdId: str
    accessToken: str
    refreshToken: str
    refreshTokenExpiry: datetime
    username: str
    mqttToken: str = None

    def __init__(self):
        """Initialize a session."""
        pass
    
    def fill(self, auth_json:str)-> None:
        self.householdId = auth_json["householdId"]
        self.accessToken = auth_json["accessToken"]
        self.refreshToken = auth_json["refreshToken"]
        self.refreshTokenExpiry = datetime.fromtimestamp(auth_json["refreshTokenExpiry"])
        self.username = auth_json["username"]
    
    def is_expired(self) -> bool:
        return self.refreshTokenExpiry

class LGHorizonPlayingInfo:
    """Represent current state of a box."""

    channel_id: str = None
    title: str = None
    image: str = None
    source_type: str = None
    paused: bool = False
    channel_title: str = None

    def __init__(self):
        """Initialize the playing info."""
        pass

    def set_paused(self, paused: bool):
        """Set pause state."""
        self.paused = paused

    def set_channel(self, channel_id):
        """Set channel."""
        self.channel_id = channel_id

    def set_title(self, title):
        """Set title."""
        self.title = title

    def set_channel_title(self, title):
        """Set channel title."""
        self.channel_title = title

    def set_image(self, image):
        """Set image."""
        self.image = image

    def set_source_type(self, source_type):
        """Set sourfce type."""
        self.source_type = source_type

    def reset(self):
        self.channel_id = None
        self.title = None
        self.image = None
        self.source_type = None
        self.paused = False
        self.channel_title = None

class LGHorizonChannel:
    """Represent a channel."""

    id: str
    title: str
    stream_image: str
    logo_image: str
    channel_number: str

    def __init__(self, channel_json):
        """Initialize a channel."""
        self.id = channel_json["id"]
        self.title = channel_json["name"]
        self.stream_image = self.get_stream_image(channel_json)
        self.logo_image = channel_json["logo"]["focused"]
        self.channel_number = channel_json["logicalChannelNumber"]
    
    def get_stream_image(self, channel_json)->str:
        image_stream = channel_json["imageStream"]
        if "full" in image_stream:
            return image_stream["full"]
        if "small" in image_stream:
            return image_stream["small"]
        logo = channel_json["logo"]
        if "focus" in logo:
            return logo["focus"]
        return ""

class LGHorizonReplayEvent:

    episodeNumber: int = None
    channelId: str = None
    eventId: str = None
    seasonNumber:int = None
    title:str = None
    episodeName:str = None

    def __init__(self, raw_json:str):
        self.channelId = raw_json["channelId"]
        self.eventId = raw_json["eventId"]
        self.title = raw_json["title"]
        if "episodeName" in raw_json:
            self.episodeName = raw_json["episodeName"]
        if "episodeNumber" in raw_json:
            self.episodeNumber = raw_json["episodeNumber"]
        if "seasonNumber" in raw_json:
            self.seasonNumber = raw_json["seasonNumber"]


class LGHorizonBaseRecording:
    
    id:str = None
    title:str = None
    image:str = None
    type:str = None
    channelId: str = None
    def __init__(self, id:str, title:str, image:str, channelId: str, type:str) -> None:
        self.id = id
        self.title = title
        self.image = image
        self.channelId = channelId
        self.type = type

class LGHorizonRecordingSingle(LGHorizonBaseRecording):
    """Represents a single recording."""
    seasonNumber: int = None
    episodeNumber: int = None

    def __init__(self, recording_json):
        """Init the single recording."""
        LGHorizonBaseRecording.__init__(
            self,
            recording_json["id"],
            recording_json["title"],
            recording_json["poster"]["url"],
            recording_json["channelId"],
            recording_json["type"]  
            )
        if "seasonNumber" in recording_json:
            self.seasonNumber = recording_json["seasonNumber"]
        if "episodeNumber" in recording_json:
            self.episodeNumber = recording_json["episodeNumber"]

class LGHorizonRecordingEpisode:
    """Represents a single recording."""
    
    episodeId: str = None
    episodeTitle:str = None
    seasonNumber: int = None
    episodeNumber: int = None
    showTitle:str = None
    recordingState:str = None
    image:str = None

    def __init__(self, recording_json):
        """Init the single recording."""
        self.episodeId = recording_json["episodeId"]
        self.episodeTitle = recording_json["episodeTitle"]
        self.showTitle = recording_json["showTitle"]
        self.recordingState = recording_json["recordingState"]
        if "seasonNumber" in recording_json:
            self.seasonNumber = recording_json["seasonNumber"]
        if "episodeNumber" in recording_json:
            self.episodeNumber = recording_json["episodeNumber"]
        if "poster" in recording_json:
            self.image = recording_json["poster"]["url"]

class LGHorizonRecordingShow:
    """Represents a single recording."""
    
    episodeId: str = None
    showTitle:str = None
    seasonNumber: int = None
    episodeNumber: int = None
    recordingState:str = None
    image:str = None

    def __init__(self, recording_json):
        """Init the single recording."""
        self.episodeId = recording_json["episodeId"]
        self.showTitle = recording_json["showTitle"]
        self.recordingState = recording_json["recordingState"]
        if "seasonNumber" in recording_json:
            self.seasonNumber = recording_json["seasonNumber"]
        if "episodeNumber" in recording_json:
            self.episodeNumber = recording_json["episodeNumber"]
        if "poster" in recording_json:
            self.image = recording_json["poster"]["url"]


class LGHorizonRecordingListSeasonShow(LGHorizonBaseRecording):
    showId:str = None
    def __init__(self, recording_season_json):
        """Init the single recording."""
        
        LGHorizonBaseRecording.__init__(
            self,
            recording_season_json["id"],
            recording_season_json["title"],
            recording_season_json["poster"]["url"],
            recording_season_json["channelId"],
            recording_season_json["type"]
            )
        if self.type == RECORDING_TYPE_SEASON:
            self.showId = recording_season_json["showId"]
        else:
            self.showId = recording_season_json["id"]

class LGHorizonVod:
    title:str = None
    image: str = None
    def __init__(self, vod_json) -> None:
        self.title = vod_json['title']

class LGHorizonApp:
    title:str = None
    image:str = None
    def __init__(self, app_state_json:str)->None:
        self.title = app_state_json["appName"]
        self.image = app_state_json["logoPath"]
        if not self.image.startswith("http:"):
                self.image = "https:" + self.image

class LGHorizonMqttClient:
    _brokerUrl:str = None
    _mqtt_client :mqtt.Client
    _auth: LGHorizonAuth
    clientId: str = None
    _on_connected_callback: Callable = None
    _on_message_callback: Callable[[str],None] = None

    @property
    def is_connected(self):
        return self._mqtt_client.is_connected

    def __init__(self, auth:LGHorizonAuth, country_settings:Dict[str,Any], on_connected_callback:Callable = None, on_message_callback:Callable[[str],None] = None):
        self._auth = auth
        self._brokerUrl = country_settings["mqtt_url"]
        self.clientId = make_id()
        self._mqtt_client = mqtt.Client(self.clientId, transport="websockets")
        self._mqtt_client.username_pw_set(self._auth.householdId, self._auth.mqttToken)
        self._mqtt_client.tls_set()
        self._mqtt_client.enable_logger(_logger)
        self._mqtt_client.on_connect = self._on_mqtt_connect
        self._on_connected_callback = on_connected_callback
        self._on_message_callback = on_message_callback
    
    def _on_mqtt_connect(self, client, userdata, flags, resultCode):
        if resultCode == 0:
            self._mqtt_client.on_message = self._on_client_message
            self._mqtt_client.subscribe(self._auth.householdId)
            self._mqtt_client.subscribe(self._auth.householdId + "/#")
            self._mqtt_client.subscribe(self._auth.householdId + "/" + self.clientId)
            self._mqtt_client.subscribe(self._auth.householdId + "/+/status")
            self._mqtt_client.subscribe(self._auth.householdId + "/+/networkRecordings")
            self._mqtt_client.subscribe(self._auth.householdId + "/+/networkRecordings/capacity")
            self._mqtt_client.subscribe(self._auth.householdId + "/watchlistService")
            self._mqtt_client.subscribe(self._auth.householdId + "/purchaseService")
            self._mqtt_client.subscribe(self._auth.householdId + "/personalizationService")
            self._mqtt_client.subscribe(self._auth.householdId + "/recordingStatus")
            self._mqtt_client.subscribe(self._auth.householdId + "/recordingStatus/lastUserAction")
            if self._on_connected_callback:
                self._on_connected_callback()
        elif resultCode == 5:
            self._mqtt_client.username_pw_set(self._auth.householdId, self._auth.mqttToken)
            self.connect()
        else:
            _logger.error(f"Cannot connect to MQTT server with resultCode: {resultCode}")
    
    def connect(self) -> None:
        self._mqtt_client.connect(self._brokerUrl, 443)
        self._mqtt_client.loop_start()
    
    def _on_client_message(self, client, userdata, message):
        """Handle messages received by mqtt client."""
        _logger.debug(f"Received MQTT message. Topic: {message.topic}")
        jsonPayload = json.loads(message.payload)
        _logger.debug(f"Message: {jsonPayload}")
        if self._on_message_callback:
            self._on_message_callback(jsonPayload)

    def publish_message(self, topic:str, json_payload:str) -> None:
        self._mqtt_client.publish(topic, json_payload)
        
    def disconnect(self) -> None:
        if self._mqtt_client.is_connected:
            self._mqtt_client.disconnect()

class LGHorizonBox:

    deviceId:str = None
    hashedCPEId:str = None
    deviceFriendlyName:str = None
    state: str = None
    playing_info: LGHorizonPlayingInfo = LGHorizonPlayingInfo()
    manufacturer:str = None
    model: str = None
    
    _mqtt_client:LGHorizonMqttClient
    _change_callback: Callable = None
    _auth: LGHorizonAuth = None
    _channels:Dict[str, LGHorizonChannel] = None
    _message_stamp = None
    
    def __init__(self, box_json:str, platform_type:Dict[str,str], mqtt_client:LGHorizonMqttClient, auth:LGHorizonAuth, channels:Dict[str, LGHorizonChannel]):
        self.deviceId = box_json["deviceId"]
        self.hashedCPEId = box_json["hashedCPEId"]
        self.deviceFriendlyName = box_json["settings"]["deviceFriendlyName"]
        self._mqtt_client = mqtt_client
        self._auth = auth
        self._channels = channels
        if platform_type:
            self.manufacturer = platform_type["manufacturer"]
            self.model = platform_type["model"]
        
    def register_mqtt(self)->None:
        if not self._mqtt_client.is_connected:
            raise Exception("MQTT client not connected.")
        topic = f"{self._auth.householdId}/{self._mqtt_client.clientId}/status"
        payload = {
            "source": self._mqtt_client.clientId,
            "state": ONLINE_RUNNING,
            "deviceType": "HGO",
        }   
        self._mqtt_client.publish_message(topic, json.dumps(payload))
    
    def set_callback(self, change_callback:Callable) -> None:
        self._change_callback = change_callback

    def update_state(self, payload):
        """Register a new settop box."""
        state = payload["state"]
        if self.state == state:
            return
        self.state = state
        if state == ONLINE_STANDBY:
            self.playing_info.reset()
            if self._change_callback:
                self._change_callback(self.deviceId)
        else:
            self._request_settop_box_state()
    
    def update_with_replay_event(self, source_type: str, event:LGHorizonReplayEvent, channel: LGHorizonChannel) -> None:
        self.playing_info.set_source_type(source_type)
        self.playing_info.set_channel(channel.id)
        self.playing_info.set_channel_title(channel.title)
        title = event.title
        if event.episodeName:
            title += f": {event.episodeName}"
        self.playing_info.set_title(title)
        self.playing_info.set_image(channel.stream_image)
        self._trigger_callback()

    def update_with_recording(self, source_type: str, recording:LGHorizonRecordingSingle, channel: LGHorizonChannel) -> None:
        self.playing_info.set_source_type(source_type)
        self.playing_info.set_channel(channel.id)
        self.playing_info.set_channel_title(channel.title)
        self.playing_info.set_title(f"{recording.title}")
        self.playing_info.set_image(recording.image)
        self._trigger_callback()
    
    def update_with_vod(self, source_type: str, vod:LGHorizonVod) -> None:
        self.playing_info.set_source_type(source_type)
        self.playing_info.set_channel(None)
        self.playing_info.set_channel_title(None)
        self.playing_info.set_title(vod.title)
        self.playing_info.set_image(None)
        self._trigger_callback()
    
    def update_with_app(self, source_type: str, app:LGHorizonApp) -> None:
        self.playing_info.set_source_type(source_type)
        self.playing_info.set_channel(None)
        self.playing_info.set_channel_title(None)
        self.playing_info.set_title(app.title)
        self.playing_info.set_image(app.image)
        self._trigger_callback()
        
    def _trigger_callback(self):
        if self._change_callback:
            _logger.debug(f"Callback called from box {self.deviceId}")
            self._change_callback(self.deviceId)

    def turn_on(self) -> None:
        """Turn the settop box on."""
        
        if self.state == ONLINE_STANDBY:
            self.send_key_to_box(MEDIA_KEY_POWER)

    def turn_off(self) -> None:
        """Turn the settop box off."""
        if self.state == ONLINE_RUNNING:
            self.send_key_to_box(MEDIA_KEY_POWER)
            self.playing_info.reset()

    def pause(self) -> None:
        """Pause the given settopbox."""
        if self.state == ONLINE_RUNNING and not self.playing_info.paused:
            self.send_key_to_box(MEDIA_KEY_PLAY_PAUSE)
    
    def play(self) -> None:
        """Resume the settopbox."""
        if self.state == ONLINE_RUNNING and self.playing_info.paused:
            self.send_key_to_box(MEDIA_KEY_PLAY_PAUSE)

    def stop(self) -> None:
        """Stop the settopbox."""
        if self.state == ONLINE_RUNNING:
            self.send_key_to_box(MEDIA_KEY_STOP)

    def next_channel(self):
        """Select the next channel for given settop box."""
        if self.state == ONLINE_RUNNING:
            self.send_key_to_box(MEDIA_KEY_CHANNEL_UP)

    def previous_channel(self) -> None:
        """Select the previous channel for given settop box."""
        if self.state == ONLINE_RUNNING:
            self.send_key_to_box(MEDIA_KEY_CHANNEL_DOWN)

    def press_enter(self) -> None:
        """Press enter on the settop box."""
        if self.state == ONLINE_RUNNING:
            self.send_key_to_box(MEDIA_KEY_ENTER)

    def rewind(self) -> None:
        """Rewind the settop box."""
        if self.state == ONLINE_RUNNING:
            self.send_key_to_box(MEDIA_KEY_REWIND)

    def fast_forward(self) -> None:
        """Fast forward the settop box."""
        if self.state == ONLINE_RUNNING:
            self.send_key_to_box(MEDIA_KEY_FAST_FORWARD)

    def record(self):
        """Record on the settop box."""
        if self.state == ONLINE_RUNNING:
            self.send_key_to_box(MEDIA_KEY_RECORD)

    def is_available(self) -> bool:
        """Return the availability of the settop box."""
        return self.state == ONLINE_RUNNING or self.state == ONLINE_STANDBY

    def set_channel(self, source:str) -> None:
        """Change te channel from the settopbox."""
        channel = [src for src in self._channels.values() if src.title == source][0]
        payload = (
            '{"id":"'
            + make_id(8)
            + '","type":"CPE.pushToTV","source":{"clientId":"'
            + self._mqtt_client.clientId
            + '","friendlyDeviceName":"Home Assistant"},'
            + '"status":{"sourceType":"linear","source":{"channelId":"'
            + channel.id
            + '"},"relativePosition":0,"speed":1}}'
        )

        self._mqtt_client.publish_message(f"{self._auth.householdId}/{self.deviceId}", payload)

    def play_recording(self, recordingId):
        """Play recording."""
        payload = (
            '{"id":"'
            + make_id(8)
            + '","type":"CPE.pushToTV","source":{"clientId":"'
            + self._mqtt_client.clientId
            + '","friendlyDeviceName":"Home Assistant"},'
            + '"status":{"sourceType":"nDVR","source":{"recordingId":"'
            + recordingId
            + '"},"relativePosition":0}}'
        )
        self._mqtt_client.publish_message(f"{self._auth.householdId}/{self.deviceId}", payload)

    def send_key_to_box(self, key: str) -> None:
        """Send emulated (remote) key press to settopbox."""
        payload = (
            '{"type":"CPE.KeyEvent","status":{"w3cKey":"'
            + key
            + '","eventType":"keyDownUp"}}'
        )
        self._mqtt_client.publish_message(f"{self._auth.householdId}/{self.deviceId}", payload)

    def _set_unknown_channel_info(self) -> None:
        """Set unknown channel info."""
        _logger.warning("Couldn't set channel. Channel info set to unknown...")
        self.playing_info.set_source_type(BOX_PLAY_STATE_CHANNEL)
        self.playing_info.set_channel(None)
        self.playing_info.set_title("No information available")
        self.playing_info.set_image(None)
        self.playing_info.set_paused(False)

    def _request_settop_box_state(self) -> None:
        """Send mqtt message to receive state from settop box."""
        topic = f"{self._auth.householdId}/{self.deviceId}"
        payload = {
            "id": make_id(8),
            "type": "CPE.getUiStatus",
            "source": self._mqtt_client.clientId,
        }
        self._mqtt_client.publish_message(topic, json.dumps(payload))

class LGHorizonCustomer:
    customerId:str = None
    hashedCustomerId:str = None 
    countryId: str = None
    cityId: int = 0
    settop_boxes: Dict[str, LGHorizonBox] = None

    def __init__(self, json_payload):
        self.customerId = json_payload["customerId"]
        self.hashedCustomerId = json_payload["hashedCustomerId"]
        self.countryId = json_payload["countryId"]
        self.cityId = json_payload["cityId"]
        if not "assignedDevices" in json_payload:
            return


